// njs_chisquare.h

#ifndef INCLUDED_NJS_CHISQUARE
#define INCLUDED_NJS_CHISQUARE

#include <njs_system.h>
#include <njs_function.h>

/*****************************************************************************
 
**************************** CHI-SQUARE ROUTINES *****************************
  
*****************************************************************************/

namespace Njs {

   namespace ChiSquare {               // ChiSquare distribution 

      double rightTail   // Chi Square right tail 
	      (double df, double chisquare, int *error, double rtol = Function::RTOL);

      inline double leftTail  // Chi Square leftTail
	      (double df, double chisquare, int *error, double rtol = Function::RTOL)
      {return *error == 0 ? 1.0 - rightTail (df, chisquare, error, rtol) : 0.0;}

      double sigLevel    // Chi Square significance level 
	      (double df, double rightTail, int *error, double rtol = Function::RTOL);

      inline double quantal     // Chi Square quantal 
	      (double df, double p, int *error, double rtol = Function::RTOL)
      {return *error == 0 ? sigLevel (df, 1.0 - p, error, rtol) : 0.0;}

      template <typename InputIterator> // vector iterators for p-values 
      double omnibusPValue (
      InputIterator begin_, // start of p-values [begin_, end_ - 1]
      InputIterator end_, // one past the last of p-values [begin_, end_ - 1]
      int *error_) // ? has a numerical error occurred : 0 = alright ?
      { // calculates the Fisher inverse chi-square (omnibus) statistic
         double sum = 0.0;
         for (InputIterator i = begin_; i != end_; i++) {
            assert (0.0 <= *i && *i <= 1.0); // probabilities are in [0.0, 1.0]
            if (*i == 0.0) return 0.0; // always significant
            sum += log (*i);
         }
         sum *= -2.0;
         // calculate the chi-square p-value
         *error_ = 0;
      return rightTail (2.0 * static_cast <double> (end_ - begin_), sum, error_);
      }

      template <typename InputIterator> // vector iterators for p-values 
      double omnibus_QFAST_PValue (
      InputIterator begin_, // start of p-values [begin_, end_ - 1]
      InputIterator end_) // one past the last of p-values [begin_, end_ - 1]
      { // calculate the Fisher inverse chi-square (omnibus) statistic
        // TL Bailey & M Gribskov (1998) 
        // "Combining evidence using p-values: application to sequence homology searches
        // Bioinformatics 14(1): 48-54
         double product = 1.0;
         InputIterator i = 0;
         for (i = begin_; i != end_; i++) {
            assert (0.0 <= *i && *i <= 1.0); // probabilities are in [0.0, 1.0]
            if (*i == 0.0) return 0.0; // always significant
            product *= *i;
         }
         double x  = -log (product);
         // calculate the chi-square p-value with the sum
         double pValue = 0.0;
         double term = 1.0;
         double number = 1.0;
         for (i = begin_; i != end_; i++) {
            pValue += term;
            term *= x / number;
            number += 1.0;
         }
         pValue *= product;
      return pValue;
      }

   }

}

#endif //! INCLUDED_NJS_CHISQUARE

