// njs_float.h

#ifndef INCLUDED_NJS_FLOAT
#define INCLUDED_NJS_FLOAT

#ifndef INCLUDED_NJS_SYSTEM
#include <njs_system.h>
#endif

namespace Njs {

   namespace Float {

      // Constants

      const double ROOT_2        =      1.4142135623730951e+000; // square root of 2.0
      const double ROOT_10       =      3.1622776601683795e+000; // square root of 10.0

      const double PI            =      3.1415926535897931e+000; // pi = ratio circumference to diameter
      const double ROOT_PI       =      1.7724538509055159e+000; // square root of pi
      const double ROOT_2_PI     =      2.5066282746310002e+000; // square root of 2.0 * pi

      const double E             =      2.7182818284590451e+000; // base of natural logarithms
      const double LN_2          =      6.9314718055994529e-001; // natural log of 2.0
      const double LN_10         =      2.3025850929940459e+000; // natural log of 10.0
      const double LN_LN_2       =     -3.6651292058166435e-001; // log of log of 2.0

      const double EULER         =      0.5772156649015325e+000; // Euler's constant

      // Approximation

      inline bool approx (float x, float y, float eps) {return fabs (x - y) <= fabs (eps);}
      inline bool relApprox (float x, float y, float eps) {return approx (x, y, eps * y);}
      inline bool absRelApprox (float x, float y, float tol, float rtol) {return approx (x, y, tol) || relApprox (x, y, rtol);}

      // Rounding

      const float FLT_THRESHOLD =       10.0F; // Rounding threshold
      const float FLT_ROUND  =          FLT_THRESHOLD * FLT_EPSILON; // Rounding error threshold

      inline bool eq (float x, float y) {return relApprox (x, y, FLT_ROUND);}
      inline bool ge (float x, float y) {return x - y >= -fabs (y) * FLT_ROUND;}
      inline bool gt (float x, float y) {return x - y >   fabs (y) * FLT_ROUND;}
      inline bool ne (float x, float y) {return ! eq (x, y);}
      inline bool lt (float x, float y) {return ! ge (x, y);}
      inline bool le (float x, float y) {return ! gt (x, y);}

      // Function Definitions

      inline double bitsToNats (double x) {return x * LN_2;}
      inline double natsToBits (double x) {return x / LN_2;}

      inline double exp2 (double x) {return exp (LN_2 * (x));}
      inline double log2 (double x) {return log (x) / LN_2;}
      inline double exp10 (double x) {return exp (LN_10 * (x));}

      inline double fmax (double x, double y) {return x > y ? x : y;}
      inline double fmin (double x, double y) {return x < y ? x : y;}
      inline double fplus (double x) {return fmax (x, 0.0);}
      inline double fminus (double x) {return fmax (-x, 0.0);}
      inline double signum (double x) {return x > 0.0 ? 1.0 : (x == 0.0 ? 0.0 : -1.0);}
      inline double heaviside (double x) {return x < 0.0 ? 0.0 : 1.0;}

      inline double psqrt (double x) {return fplus (sqrt (x));}
      inline double square (double x) {return x * x;}

      inline double bound (double x, double xlo, double xhi) {
         if (x < xlo) return xlo;
         if (x < xhi) return x;
         return xhi;
      }

      // Approximation

      inline bool approx (double x, double y, double eps) {return fabs (x - y) <= fabs (eps);}
      inline bool relApprox (double x, double y, double eps) {return approx (x, y, eps * y);}
      inline bool absRelApprox (double x, double y, double tol, double rtol) {return approx (x, y, tol) || relApprox (x, y, rtol);}

      // Rounding

      const double DBL_THRESHOLD =      100.0; // Rounding threshold
      const double DBL_ROUND  =         DBL_THRESHOLD * DBL_EPSILON; // Rounding error threshold

      inline bool eq (double x, double y) {return relApprox (x, y, DBL_ROUND);}
      inline bool ge (double x, double y) {return x - y >= -fabs (y) * DBL_ROUND;}
      inline bool gt (double x, double y) {return x - y >   fabs (y) * DBL_ROUND;}
      inline bool ne (double x, double y) {return ! eq (x, y);}
      inline bool lt (double x, double y) {return ! ge (x, y);}
      inline bool le (double x, double y) {return ! gt (x, y);}

      // Function Definitions

      inline float bitsToNats (float x) {return x * static_cast <float> (LN_2);}
      inline float natsToBits (float x) {return x / static_cast <float> (LN_2);}

      inline float exp2 (float x) {return static_cast <float> (exp (LN_2 * x));}
      inline float log2 (float x) {return static_cast <float> (log (x) / LN_2);}
      inline float exp10 (float x) {return static_cast <float> (exp (LN_10 * x));}

      inline float fmax (float x, float y) {return x > y ? x : y;}
      inline float fmin (float x, float y) {return x < y ? x : y;}
      inline float fplus (float x) {return fmax (x, 0.0F);}
      inline float fminus (float x) {return fmax (-x, 0.0F);}
      inline float signum (float x) {return x > 0.0F ? 1.0F : (x == 0.0F ? 0.0F : -1.0F);}
      inline float heaviside (float x) {return x < 0.0F ? 0.0F : 1.0F;}

      inline float psqrt (float x) {return static_cast <float> (fplus (sqrt (x)));}
      inline float square (float x) {return x * x;}

      inline float bound (float x, float xlo, float xhi) {
         if (x < xlo) return xlo;
         if (x < xhi) return x;
         return xhi;
      }

      inline float probability (float x) {return bound (x, 0.0F, 1.0F);}

   };

};

#endif //! INCLUDED_NJS_FLOAT
