// njs_function.cpp

#undef NJS_DEBUG

#include <njs_function.h>

#include <njs_combinatoric.h>
#include <njs_debug.h>
#include <njs_float.h>
#include <njs_int.h>
#include <njs_io.h>

using namespace Njs;

double Function::seriesSum (                 // series sum 
int m,                                     // starting term 
int n,                                     // ending term 
void *p,                                    // parameters 
double (*term_function) (int n, void *p),  // function giving term 
int *error,                                // error status 
double rtol)                                // relative accuracy 
{
   int i, incr;
	double sum = 0.0;
	double term = 0.0;
   if (*error != 0) {
return 0.0;
   }
	incr = m < n ? 1 : -1;
	for (i = m; incr * (n - i) >= 0 ; i += incr) {
		sum += (term = term_function (i, p));
      if (sum != 0.0 && fabs (term / sum) < rtol) {
			*error = 0;
return sum;
		}
	}
	*error = 1;
return sum;
}

double Function::continuedFraction (        // continued fraction limit 
int m,                                     // starting term 
int n,                                     // ending term 
void *p,                                    // parameters 
double (*a) (int n, void *p),              // function giving term 
double (*b) (int n, void *p),              // function giving term 
int *error,                                // error = 1 : premature termination of series
double rtol)                                // relative tolerance for error
{
   int i, incr;
	double ac [2], bc [2];
	double as, bs, bcc;
	double limit = 0.;
	double limitold = 0.;
   if (*error != 0) {
return 0.0;
   }
	incr = m < n ? 1 : -1;
	bc [m % 2] = 0.;
	ac [m % 2] = bc [(m + 1) % 2] = 1.;
	ac [(m + 1) % 2] = b (m - 1, p);
	for (i = m; incr * (n - i) >= 0 ; i += incr) {
		bs = b (i, p);
		as = a (i, p);
		ac [i % 2] = bs * ac [(i + 1) % 2] 
			+ as * ac [i % 2];
		bcc = bc [i % 2] = bs * bc [(i + 1) % 2] + as * bc [i % 2];
		if (bc [i % 2] != 0.) {
			ac [0] /= bcc;
			bc [0] /= bcc;
			ac [1] /= bcc;
			bc [1] /= bcc;
			limit = ac [i % 2];
         if (limit != 0.0 && Njs::Float::relApprox (limitold, limit, rtol)) {
				*error = 0;
return limit;
			}
			limitold = limit;
		}
	}
	*error = 1;
return limit;
}

// Start of static Gamma Function Routines

static double logDerivative (int n, double *u)  // values of u, u', u", etc. 
{
   if (n < 0) {
Njs::Io::abort ("static Function::logDerivative : negative derivative order is illegal");
	}
	int i;
	double *y = new double [n + 1];
	double value;
	for (i = 1; i <= n; i++) {
		y [i] = u [i] / u [0];
	}
	if (n == 0) {
		value = log (u [0]);
   } else if (n == 1) {
		value = y [1];
   } else if (n == 2) {
		value = y [2] - y [1] * y [1];
   } else if (n == 3) {
		value = y [3] - 3. * y [2] * y [1] + 2. * y [1] * y [1] * y [1];
   } else if (n == 4) {
		value = y [4] - 4. * y [3] * y [1] - 3. * y [2] * y [2] 
						  + 12. * y [2] * y [1] * y [1] 
						  - 6. * y [1] * y [1] * y [1] * y [1];
   } else {
Njs::Io::abort ("static Function::logDerivative : illegal derivative order");
	}
   delete [] y;
   y = 0;
return value;
}

static double S_COEFF [] = {
    1.000000000000000e+00,
	 4.694580336184385e+04,
	-1.560605207784446e+05,
	 2.065049568014106e+05,
	-1.388934775095388e+05,
	 5.031796415085709e+04,
	-9.601592329182778e+03,
	 8.785855930895250e+02,
	-3.155153906098611e+01,
	 2.908143421162229e-01,
	-2.319827630494973e-04,
	 1.251639670050933e-10};

static double *S_COF   = S_COEFF;
static int S_GAMMA_SERIES = 11;
static double S_GAMMA    = 11.;

void Function::setGammaCoeff (double *cof, int gammaSeries, double gamma)
{
   S_COF = cof;
	S_GAMMA_SERIES = gammaSeries;
	S_GAMMA = gamma;
}

static double seriesLnGamma (
double x, 
int n)                                    // order of the derivative 
{
NJS_MSG("+ seriesLnGamma");
	double xx, tmp, value;
	double *y = new double [n];
	xx = x - 1.;                            // normalize from gamma (x + 1) to xx ! 
	for (int k = 0; k < n; k++) {
	   y [k] = 0.;
	   for (int i = S_GAMMA_SERIES; 1 <= i; i--) {
			y [k] += S_COF [i] / Int::integerPower <double, int> (xx + i, k + 1);
		}
      y [k] *= Int::minusOnePower <int> (k) * Combinatoric::factorial (k);
	}
   y [0] += S_COF [0];
	value = logDerivative (n - 1, y);
	delete [] y;
   y = 0;
	tmp = xx + S_GAMMA + 0.5;
	if (n == 1) {
      value += log (Float::ROOT_2_PI) + (xx + 0.5) * log (tmp) - tmp;
	} else if (n == 2) {
		value += log (tmp) - S_GAMMA / tmp;
	} else {
		value += Int::minusOnePower <int> (n - 1) * Combinatoric::factorial (n - 3)
				 * Int::integerPower <double, int> (tmp, 2 - n)
			    *	(1. + (n - 2) * S_GAMMA / tmp);
	}
return value;
}

static double generalLnGamma(double x, int n)  
// order of the derivative 
// n = 1, 2, ... ln(gamma), digamma, ...    
{
NJS_MSG("+ static Function::generalLnGamma")
	double value;
	double *y;
	if (x < 0.0) {
		value = Int::minusOnePower <int> (n - 2) * seriesLnGamma (1.0 - x, n);
		if (n == 1) {
		   value += log (Float::PI) - log (fabs (sin (Float::PI * x)));
		} else {
	      y = new double [n];
			for (int k = 0; k < n; k++) {
				y [k] = Int::integerPower (Float::PI, k) * sin (Float::PI * x + Float::PI / 2.0 * k);
			}
			value += -logDerivative (n - 1, y);
	      delete [] y;
         y = 0;
		}
	} else if (0.0 <= x && x < 1.0) {
		value = seriesLnGamma (1.0 + x, n);
		if (n == 1) {
		   value += -log (x);
		} else {
         value += Int::minusOnePower <int> (n - 1) * Combinatoric::factorial <int> (n - 2) 
					 * Int::integerPower (x, 1 - n);
		}
	} else if (1.0 <= x) {
return seriesLnGamma (x, n);
	}
return value;
} 

// End of static Gamma Function Routines
// Start of Gamma Function Routines

double Function::factorial (double x)
{
NJS_MSG("+ Function::factorial")
   if (x <= LONG_MIN) {
Njs::Io::abort ("Function::factorial : x <= LONG_MIN");
   }
   double sign = -1.0 <= x ? 1.0 : Int::minusOnePower <int> ((int) (-x));
return sign * exp (generalLnGamma (x + 1.0, 1));
} 

double Function::lnGamma (double x)
{
return generalLnGamma (x, 1);
} 

double Function::diGamma (double x)
{
return generalLnGamma (x, 2);
} 

double Function::triGamma (double x)
{
return generalLnGamma (x, 3);
} 

// End of Gamma Function Routines

static bool logPGammaInit = false;

static double pGammaTerm (int n, void *p)
{
	static double term;
	double x = ((double *) p) [0];
   double a = ((double *) p) [1];
	if (logPGammaInit) {
		logPGammaInit = false;
      term = Int::integerPower <double, int> (x, n) / exp (Function::lnGamma (a + 1.0 + n));
	} else {
		term *= x / (a + n);
	}
return term;
}

static double pGammaSeries (double a, double x, int *error, double rtol)
{
   double p [] = {x, a};
	const int ITMAX = 1000;
   if (*error != 0) {
return 0.0;
   }
	logPGammaInit = true;
	double sum = exp (-x + a * log(x)) * 
      Function::seriesSum (0, ITMAX, (void *) p, pGammaTerm, error, rtol);
return sum;
}

static double qGammaA (int n, void *p)
{
	if (n % 2 == 0) {
return (n / 2) - ((double *) p) [1];
	} else if (n == 1) {
return 1.0;
	} else {
return (double) (n / 2);
	}
}

static double qGammaB (int n, void *p)
{
	if (n == 0) {
return 0.;
	} else if (n % 2 == 0) {
return 1.;
	} else {
return ((double *) p) [0];
	}
}

static double qGammaFraction (double a, double x, int *error, double rtol)
{
//
// Formula calculated according to Eq. (6.5.31), p. 263 in
// M. Abramowitz & I.A. Stegun (1965) "Handbook of Mathematical Functions"
// Dover, New York
//
   double p [] = {x, a};
	const int ITMAX = 1000;
   if (*error != 0) {
return 0.0;
   }
   double sum = exp (-x + a * log(x)) / exp (Function::lnGamma (a)) * 
      Function::continuedFraction (1, ITMAX, p, qGammaA, qGammaB, error, rtol);
return sum;
}

double Function::incompleteQGamma (        // Incomplete Gamma Function Q(a,x) 
double a,
double x,
int *error,
double rtol)
{
	if (x < 0.0) {
Njs::Io::abort ("incompleteQGamma : x < 0.0");
	} else if (a <= 0.0) {
Njs::Io::abort ("incompleteQGamma : a <= 0.0");
	}
   if (*error != 0) {
return 0.0;
   }
	if (x == 0.0) {
return 1.0;		
	} else if (x < a + 1.0) {
return 1.0 - pGammaSeries (a, x, error, rtol);
	} else {
return qGammaFraction (a, x, error, rtol);
	}
}

double Function::erf (                     // Error Function 
double x,
int *error,
double rtol)
{
   if (*error != 0) {
return 0.0;
   }
return Float::signum (x) * (1.0 - incompleteQGamma (0.5, x * x, error, rtol));
}

double Function::erfc (                    // Complementary Error Function 
double x,
int *error,
double rtol)
{
   if (*error != 0) {
return 0.0;
   }
return incompleteQGamma (0.5, x * x, error, rtol);
}

static double millsRatioA (int n, void *p)
{
	if (n == 1) {
return 1.0;
	} else {
return (double) (n - 1.0);
	}
}

static double millsRatioB (int n, void *p)
{
	if (n == 0) {
return 0.0;
	} else {
return * ((double *) p);
	}
}

double Function::millsRatio (                  // (1 - Phi (x)) / phi (x) 
double x,
int *error,
double rtol)
{ // ratio of normal tail to normal density 
	const int ITMAX = 1000;
   if (*error != 0) {
return 0.0;
   }
return Function::continuedFraction (1, ITMAX, (void *) &x, 
			millsRatioA, millsRatioB, error, rtol);
}

double Function::oneMinusExp (double x) // 1.0 - exp (x) 
{
   const double EPS = 1.0e-20;
   if (x == 0.0) return 0.0;
   if (fabs (x) > 1.0) return 1.0 - exp (x);
   long n = 0;
   double term = 1.0;
   double term0 = 1.0;
   do {
      term0 = term;
      n++;
      term *= x / static_cast <double> (n);
   } while (fabs (term / x) > EPS);
   n--;
   term = term0;
// term holds the last non-negligible term, which is pow (x, n) / n!
// sum is in reverse order for accuracy
   double sum = 0.0;
   do {
      sum += term;
      term *= static_cast <double> (n) / x;
      n--;
   } while (n > 0.0);
return -sum;
}


