// njs_seqread.h

#ifndef INCLUDED_NJS_SEQREAD
#define INCLUDED_NJS_SEQREAD

#include <njs_system.h>
#include <njs_seqreadproto.h>

/*****************************************************************************
 
************************* SEQUENCE READING ROUTINES **************************
  
*****************************************************************************/

namespace Njs {

   typedef std::vector <char **> CharPtrPtrVec;

   class SeqRead : public SeqReadProto, protected CharPtrPtrVec {

      public:

      SeqRead (bool storeDefLine_ = false) : d_storeDefLine (storeDefLine_), d_strlen (0) {}

      virtual ~SeqRead () {free ();}

      virtual operator bool () const {return size () != 0;} // ? does the SeqRead contain something?

      virtual SeqReadProto *addFastaFile (
      std::ifstream &inFileStream_, // input filestream : input in fasta format
      const char *buffer_, // sequence buffer
      size_t dimBuffer_); // dimension of the sequence buffer (including trailing '\0')
      // buffer_ MUST be able to hold the longest line in the input

      virtual SeqReadProto *addFastaFile (
      const char *inFileName_, // filename : input in fasta format
      const char *buffer_, // sequence buffer
      size_t dimBuffer_); // dimension of the sequence buffer (including trailing '\0')
      // buffer_ MUST be able to hold the longest line in the input

      virtual SeqReadProto *free (); // clears the sequences and frees memory.

      virtual std::ostream &out (std::ostream &ostr_) const; // output.

      void push_back (char *seq_); // pushes back a copy of a new sequence with blank defline
	   // for DEBUGGING only
      void push_back (char **seq_); // pushes back a copy of a defline and sequence
	   // for DEBUGGING only

	   virtual inline SeqReadProto *initialize (HashProto *hash_) {return this;} // performs initialization of the object after sequences are loaded
      virtual inline SeqReadProto *setStoreDefLine (bool storeDefLine_) {d_storeDefLine = storeDefLine_; return this;} // clears the sequences and frees memory.
      virtual inline char *getSeq (size_t i) const {return CharPtrPtrVec::at (i) [1];} // pointer to the 0-th char in seq i
      virtual inline char *getSeqs (size_t i) const {return getSeq (i);} // pointer to the 0-th char in seq i
	   virtual inline size_t getStrLen (size_t i) const {return d_strlen [i];} // returns the strlen of seq i : access to d_strlen, not calculation

      inline bool getStoreDefLine () const {return d_storeDefLine;} // ? store the def-line ?
      inline const char *getDefLine (size_t i) const {return CharPtrPtrVec::at (i) [0];} // pointer to the 0-th char in def-line i
	   inline const size_t getCountSeqs () const {return CharPtrPtrVec::size ();} // pointer to the number of sequences in the file.
       virtual std::vector <char **>::size_type size () const {return CharPtrPtrVec::size ();} // extension of the std::vector <char **> function
      virtual std::vector <char **>::const_iterator begin () const {return CharPtrPtrVec::begin ();} // extension of the std::vector <char **> function
      virtual std::vector <char **>::const_iterator end () const {return CharPtrPtrVec::end ();} // extension of the std::vector <char **> function

      private:

      bool d_storeDefLine; // ? store the def-line ?
      std::vector <size_t> d_strlen; // stores the string lengths

      SeqRead &operator= (const SeqRead &seqRead_); // disallow copying

      SeqReadProto *addFastaBuffer (
      const char *buffer_, // buffer of characters
      bool isDefLine_ = false); // ? Is the buffer a defline ? 

   };

    class SeqReadOut {

      public:

      static SeqReadOut *out;

      inline static void set (SeqReadOut *out_) {out = out_;} // sets out state
      static void clear (); // resets out state to default

      inline SeqReadOut () : outLength (false) {set (this);}

      inline ~SeqReadOut () {clear ();}

      bool outLength;

   };

}

inline std::ostream &operator<< (std::ostream &ostr_, const Njs::SeqRead &seqRead_)
{return seqRead_.out (ostr_);}

#endif //! INCLUDED_NJS_SEQREAD











