// njs_variate.h

#ifndef INCLUDED_NJS_VARIATE
#define INCLUDED_NJS_VARIATE

#ifndef INCLUDED_NJS_SYSTEM
#include <njs_system.h>
#endif

#ifndef INCLUDED_NJS_IO
#include <njs_io.h>
#endif
#ifndef INCLUDED_NJS_MEMUTIL
#include <njs_memutil.h>
#endif

/*****************************************************************************
 
************************** RANDOM NUMBER ROUTINES ****************************
  
*****************************************************************************/

// Random no. seeder and generator 

namespace Njs {

   namespace Variate {

      const long RANDOM_SEED = 31415; // default random seed

      const double BAD_PROBABILITY = -HUGE_VAL;

      double probability (double x);

      void randomSeed (long n);           // provides seed for random numbers

      long randomNum ();                  // returns random numbers

      template <class T>
      class Counter {                     // counter for Monte Carlo routines

         public:

         inline Counter (                 // keeps counts of bins presented 
         size_t dimBin_ = 0,              // #(distinct bins) : must be positive          
         const T *bin_ = 0,               // bin_ [dimBin_] : bins          
         const long *count_ = 0)          // count_ [dimBin_ + 1] : counts 
            : d_dimBin (0), d_bin (0), d_count (0)
         {copy (dimBin_, bin_, count_);}

         inline Counter (                 // initialization of counter 
         const Counter &counter_)         // counter
            : d_dimBin (0), d_bin (0), d_count (0)
         {copy (counter_);}

         inline ~Counter () {free ();}

         inline Counter &operator= (      // assignment of counter 
         const Counter<T> &counter_)      // counter
         {return this == &counter_ ? *this : copy (counter_);}

         inline operator bool ()          // is counter valid?
         const {return d_dimBin != 0;}

         Counter<T> &free ()              // counter  
         {
            if (d_dimBin) {
               delete [] d_count;
               d_count = 0;
               delete [] d_bin;
               d_bin = 0;
            }
            d_dimBin = 0;
         return *this;
         }

         Counter<T> &copy (               // counter  
         size_t dimBin_,                  // #(distinct bins)          
         const T *bin_,                   // bin_ [dimBin_] : bins          
         const long *count_)              // count_ [dimBin_ + 1] : counts          
         {
            if (d_dimBin != dimBin_) {
               free ();
               init (dimBin_);
            }
         // initialize
            size_t i = 0;
            if (d_dimBin) {
               for (i = 0; i < dimBin_; i++) {
                  d_bin [i] = bin_ [i];
               }
               if (count_ == 0) {
                  Njs::MemUtil::memZero (d_count, (dimBin_ + 1) * sizeof (long));
               } else for (i = 0; i < dimBin_ + 1; i++) {
                  d_count [i] = count_ [i];
               }
            }
         return *this;
         }

         inline Counter<T> &copy (         // counter  
         const Counter &counter_)          // pre-existing counter
         {return copy (counter_.d_dimBin, counter_.d_bin, counter_.d_count);}

         inline Counter<T> &zeroAllCounts () // resets all the counters
         {Njs::MemUtil::memZero (d_count, (d_dimBin + 1) * sizeof (long)); return *this;}

         inline Counter<T> &clear () // resets all the counters
         {return zeroAllCounts ();}

         inline Counter<T> &setCount (const T &value_, long count_) 
         {d_count [search (value_)] = count_; return *this;} // i-th d_bin d_count

         inline Counter<T> &incrementCountOfValue (const T &value_) // adds 1 to the counter d_bin [i - 1] < value_ < d_bin [i]
         {d_count [search (value_)]++; return *this;}

         inline long getCountOfValue (const T &value_) 
         const {return d_count [search (value_)];} // d_bin d_count for value_

         inline size_t getDimBin () const {return d_dimBin;} // #(distinct bins)
         inline const T *getBin () const {return d_bin;} // i-th element of d_bin
         inline const long *getCount () const {return d_count;} // i-th d_bin d_count

         private:

         size_t d_dimBin;                    // #(distinct bins)          
         T *d_bin;                           // bins 
         long *d_count;                      // counter of corresponding bins  

         Counter<T> &init (                   // counter  
         size_t dimBin_)                     // #(distinct bins)          
         {
            if (dimBin_) {
               try {
                  d_bin = new T [dimBin_];
                  d_count = new long [dimBin_ + 1];
               } catch (std::bad_alloc &) {
         Njs::Io::abortFormat ("Counter &Counter<T>::init : Failed Allocation");
               }
            }
            d_dimBin = dimBin_;
         return *this;
         }

         size_t search (const T &value_) // returns i : d_bin [i - 1] < value_ <= d_bin [i]
         // returns 0           if value_ <= d_bin [0]                 (sentinel -HUGE_VAL)
         // returns dimBin_     if d_bin [d_dimBin - 1] < value_       (sentinel +HUGE_VAL)
         // returns i           if d_bin [i - 1] < value_ <= d_bin [i]
         const {
            if (d_dimBin == 0) { // there is only one category
         return 0;
            } else if (value_ <= d_bin [0]) {
         return 0;
            } else if (d_bin [d_dimBin - 1] < value_) {
         return d_dimBin;
            } else { // the condition d_bin [0] < value <= d_bin [d_dimBin - 1] holds
               size_t i = 0;
               size_t j = d_dimBin - 1;
               while (j != i + 1) { // the condition d_bin [i] < value_ <= d_bin [j] holds throughout
                  size_t test = (i + j) / 2;
                  if (value_ < d_bin [test]) {
                     j = test;
                  } else if (d_bin [test] < value_) {
                     i = test;
                  } else {
         return test;
                  }
               }
         return j; // d_bin [j - 1] < value_ <= d_bin [j]
            }
         return 0; // just to make the compiler happy : this line is never reached
         }

      };

   }

}

#endif //! INCLUDED_NJS_VARIATE

